local const = require('const')
local json = require("json")
local ldbus = require("ldbus")
local utils = require("utils3")
local watcher = require("watcher2")
local shadow = require("shadow2")

local service_prefix = "iot.gateway.webconsole"

utils.check_request_method("POST")
utils.check_auth()
utils.check_permission("DeviceManage")

local function random_num_7()
    math.randomseed(tostring(os.time()):reverse():sub(1, 7))
    return math.random(1, 100)
end

local function current_time()
    ngx.update_time()
    return ngx.now() * 1000
end

local function dbus_connection_open()
    if watch_dbus_connection then
        return watch_dbus_connection, nil
    end

    local bus = "unix:path=/tmp/var/run/mbusd/mbusd_socket"
    local conn, err_o = ldbus.connection.open(bus)
    if not conn then
        return nil, string.format("open connection to %s: %s", bus, tostring(err_o))
    end

    local reg, err_r = ldbus.bus.register(conn)
    if not reg then
        return nil, "register connection: " .. tostring(err_r)
    end

    watch_dbus_connection = conn
    return conn, nil
end

local function dbus_request_name(conn, name)
    utils.log_info("request name: %s", name)
    local req, err = ldbus.bus.request_name(conn, name, {
        replace_existing = true,
        allow_replacement = true,
    })
    if req ~= "primary_owner" and req ~= "already_owner" then
        return string.format("request name %s: %s", name, err)
    end
    return nil
end

local function dbus_release_name(conn, name)
    utils.log_info("release name: %s", name)
    local req, err = ldbus.bus.release_name(conn, name)
    if req ~= "released" and req ~= "non_existant" and req ~= "not_owner" then
        return string.format("release name %s: %s", name, err)
    end
    return nil
end

local function set_properties(things)
    if not utils.is_array_table(things) or #things == 0 then
        return utils.error_exit(
                ngx.HTTP_BAD_REQUEST,
                const.INVALID_PARAMETER,
                "Things not a array or is empty")
    end

    local resp = {}
    resp["Data"] = {}
    resp["Data"]["Results"] = {}
    for _, v in pairs(things) do
        -- Handle parameters and get properties from things.
        local product = v["ProductKey"]
        if type(product) ~= "string" or #product == 0 then
            return utils.error_exit(
                    ngx.HTTP_BAD_REQUEST,
                    const.INVALID_PARAMETER,
                    "ProductKey not a string or is empty")
        end
        if not utils.is_valid_pk(product) then
            return utils.error_exit(
                    ngx.HTTP_BAD_REQUEST,
                    const.INVALID_PARAMETER,
                    "illegal ProductKey")
        end

        local device = v["DeviceName"]
        if type(device) ~= "string" or #device == 0 then
            return utils.error_exit(
                    ngx.HTTP_BAD_REQUEST,
                    const.INVALID_PARAMETER,
                    "DeviceName not a string or is empty")
        end
        if not utils.is_valid_dn(device) then
            return utils.error_exit(
                    ngx.HTTP_BAD_REQUEST,
                    const.INVALID_PARAMETER,
                    "illegal DeviceName")
        end

        local properties = v["Properties"]
        if type(properties) ~= "table" then
            return utils.error_exit(
                    ngx.HTTP_BAD_REQUEST,
                    const.INVALID_PARAMETER,
                    "Properties not a object or is empty")
        end
        local k = next(properties)
        if not k then
            return utils.error_exit(
                    ngx.HTTP_BAD_REQUEST,
                    const.INVALID_PARAMETER,
                    "Properties not a object or is empty")
        end

        local r = {}
        r["ProductKey"] = product
        r["DeviceName"] = device
        r["Returns"] = {}
        local thing_shadow = shadow:new(nil, product, device)
        local ok, res, err = thing_shadow:set_properties(properties)
        if ok then
            r["Returns"]["Data"] = res
        else
            r["Returns"]["Code"] = res
            r["Returns"]["Message"] = err
        end
        table.insert(resp["Data"]["Results"], r)
    end
    resp["Data"]["Timestamp"] = current_time()
    local str = json.encode(resp)

    ngx.header['Content-Type'] = 'application/json; charset=utf-8'
    ngx.say(str)
end

local function get_properties(things)
    if not utils.is_array_table(things) or #things == 0 then
        return utils.error_exit(
                ngx.HTTP_BAD_REQUEST,
                const.INVALID_PARAMETER,
                "Things not a array or is empty")
    end

    local resp = {}
    resp["Data"] = {}
    resp["Data"]["Results"] = {}
    for _, v in pairs(things) do
        -- Handle parameters and get properties from things.
        local product = v["ProductKey"]
        if type(product) ~= "string" or #product == 0 then
            return utils.error_exit(
                    ngx.HTTP_BAD_REQUEST,
                    const.INVALID_PARAMETER,
                    "ProductKey not a string or is empty")
        end
        if not utils.is_valid_pk(product) then
            return utils.error_exit(
                    ngx.HTTP_BAD_REQUEST,
                    const.INVALID_PARAMETER,
                    "illegal ProductKey")
        end

        local device = v["DeviceName"]
        if type(device) ~= "string" or #device == 0 then
            return utils.error_exit(
                    ngx.HTTP_BAD_REQUEST,
                    const.INVALID_PARAMETER,
                    "DeviceName not a string or is empty")
        end
        if not utils.is_valid_dn(device) then
            return utils.error_exit(
                    ngx.HTTP_BAD_REQUEST,
                    const.INVALID_PARAMETER,
                    "illegal DeviceName")
        end

        local identifiers = v["Identifiers"]
        if not utils.is_array_table(identifiers) or #identifiers == 0 then
            return utils.error_exit(
                    ngx.HTTP_BAD_REQUEST,
                    const.INVALID_PARAMETER,
                    "Identifiers not a array or is empty")
        end

        local r = {}
        r["ProductKey"] = product
        r["DeviceName"] = device
        r["Returns"] = {}
        local thing_shadow = shadow:new(nil, product, device)
        local ok, properties, err = thing_shadow:get_properties(identifiers)
        if ok then
            r["Returns"]["Data"] = properties
        else
            r["Returns"]["Code"] = properties
            r["Returns"]["Message"] = err
        end
        table.insert(resp["Data"]["Results"], r)
    end
    resp["Data"]["Timestamp"] = current_time()
    local str = json.encode(resp)

    ngx.header['Content-Type'] = 'application/json; charset=utf-8'
    ngx.say(str)
end

local function call_services(things)
    if not utils.is_array_table(things) or #things == 0 then
        return utils.error_exit(
                ngx.HTTP_BAD_REQUEST,
                const.INVALID_PARAMETER,
                "Things not a array or is empty")
    end

    local resp = {}
    resp["Data"] = {}
    resp["Data"]["Results"] = {}
    for _, v in pairs(things) do
        -- Handle parameters and get properties from things.
        local product = v["ProductKey"]
        if type(product) ~= "string" or #product == 0 then
            return utils.error_exit(
                    ngx.HTTP_BAD_REQUEST,
                    const.INVALID_PARAMETER,
                    "ProductKey not a string or is empty")
        end
        if not utils.is_valid_pk(product) then
            return utils.error_exit(
                    ngx.HTTP_BAD_REQUEST,
                    const.INVALID_PARAMETER,
                    "illegal ProductKey")
        end

        local device = v["DeviceName"]
        if type(device) ~= "string" or #device == 0 then
            return utils.error_exit(
                    ngx.HTTP_BAD_REQUEST,
                    const.INVALID_PARAMETER,
                    "DeviceName not a string or is empty")
        end
        if not utils.is_valid_dn(device) then
            return utils.error_exit(
                    ngx.HTTP_BAD_REQUEST,
                    const.INVALID_PARAMETER,
                    "illegal DeviceName")
        end

        local services = v["Services"]
        if not utils.is_array_table(services) or #services == 0 then
            return utils.error_exit(
                    ngx.HTTP_BAD_REQUEST,
                    const.INVALID_PARAMETER,
                    "Services not a array or is empty")
        end

        local r = {}
        r["ProductKey"] = product
        r["DeviceName"] = device
        r["Services"] = {}
        local thing_shadow = shadow:new(nil, product, device)
        for _, s in pairs(services) do
            local name = s["Name"]
            if type(name) ~= "string" or #name == 0 then
                return utils.error_exit(
                        ngx.HTTP_BAD_REQUEST,
                        const.INVALID_PARAMETER,
                        "Service.Name not a string or is empty")
            end
            local args = s["Args"]

            local m = {}
            m["Name"] = name
            m["Returns"] = {}
            local ok, ret, err = thing_shadow:call_service(name, args)
            if ok then
                m["Returns"]["Data"] = ret
            else
                m["Returns"]["Code"] = ret
                m["Returns"]["Message"] = err
            end
            table.insert(r["Services"], m)
        end
        table.insert(resp["Data"]["Results"], r)
    end
    resp["Data"]["Timestamp"] = current_time()
    local str = json.encode(resp)

    ngx.header['Content-Type'] = 'application/json; charset=utf-8'
    ngx.say(str)
end

local function watch_heartbeat_resp()
    local res = {}
    res["Message"] = "heartbeat"
    res["Data"] = {}
    res["Data"]["Timestamp"] = current_time()
    return json.encode(res)
end

local function watch_message_resp(msg)
    if not msg then return nil end

    local event_name = msg:get_member()
    local path = msg:get_path()
    local product, device = string.match(path, "/sys/(.+)/(.+)/thing/event/.+/post")
    if not product or not device then
        utils.log_info("ignore object path: %s", path)
        return nil
    end

    local iter = ldbus.message.iter.new()
    if not msg:iter_init(iter) then
        utils.log_info("iter init failed")
        return nil
    end

    if iter:get_arg_type() ~= ldbus.types.string then
        utils.log_info("arg type not a string")
        return nil
    end

    local res = {}
    res["Message"] = "success"
    res["Data"] = {}
    res["Data"]["ProductKey"] = product
    res["Data"]["DeviceName"] = device
    res["Data"]["Event"] = event_name
    res["Data"]["Args"] = nil
    res["Data"]["Timestamp"] = current_time()

    local args_str = iter:get_basic()
    if args_str then
        local ok, args = pcall(json.decode, args_str)
        if not ok then
            utils.log_info("decode message failed: " .. args_str)
            return nil
        end
        res["Data"]["Args"] = args
    end

    return json.encode(res)
end

local function watch(things)
    local conn, err_c = dbus_connection_open()
    if not conn then
        return utils.error_exit(
                ngx.HTTP_INTERNAL_SERVER_ERROR,
                const.CONNECTION_CREATE,
                err_c)
    end

    local service_name = service_prefix .. ngx.worker.pid() .. random_num_7()
    local err_n = dbus_request_name(conn, service_name)
    if err_n then
        return utils.error_exit(
                ngx.HTTP_INTERNAL_SERVER_ERROR,
                const.CONNECTION_CREATE,
                err_n)
    end

    local wtr = watcher:new(nil, conn, service_name)
    -- The only error in generate_watch_rule is invalid request message framing.
    local rule, err_r = wtr:generate_watch_rule(things, service_name)
    if not rule then
        dbus_release_name(conn, service_name)
        return utils.error_exit(
                ngx.HTTP_BAD_REQUEST,
                const.INVALID_PARAMETER,
                err_r)
    end

    local err_w = wtr:add_watch(rule)
    if err_w then
        dbus_release_name(conn, service_name)
        return utils.error_exit(
                ngx.HTTP_INTERNAL_SERVER_ERROR,
                const.MESSAGE_SENDANDREPLY,
                err_w)
    end

    ngx.header['Content-Type'] = 'text/plain; charset=utf-8'
    local count, ok, abort = 0, false, false
    local msg, resp, err_f
    while conn:read_write(1 * 1000) do
        repeat
            msg = conn:pop_message()
            resp = watch_message_resp(msg)
            if resp then ngx.say(resp) end

            count = count + 1
            if count >= 30 then
                count = 0
                resp = watch_heartbeat_resp()
                ngx.say(resp)
            end

            if resp then
                ok, err_f = ngx.flush(true)
                if not ok then
                    utils.log_warn("cannot write to client: %s", err_f)
                    abort = true
                    break
                end
            end
        until not msg
        if abort then
            break
        end
    end
    wtr:remove_watch(rule)
    dbus_release_name(conn, service_name)
    -- FIXME How to drain the incoming queue?
    if not abort then
        return utils.error_exit(
                ngx.HTTP_INTERNAL_SERVER_ERROR,
                const.CONNECTION_LOST,
                "internal connection is disconnected")
    end
end

local bulk_action_set_properties = "setproperties"
local bulk_action_get_properties = "getproperties"
local bulk_action_call_services = "callservices"
local bulk_action_watch = "watch"

local bulk_action_switch = {
    [bulk_action_get_properties] = get_properties,
    [bulk_action_set_properties] = set_properties,
    [bulk_action_call_services] = call_services,
    [bulk_action_watch] = watch,
}

ngx.req.read_body()
local body = ngx.req.get_body_data()
if not body then
    return utils.error_exit(
            ngx.HTTP_BAD_REQUEST,
            const.INVALID_PARAMETER,
            "no body found")
end

local ok, args = pcall(json.decode, body)
if not ok then
    return utils.error_exit(
            ngx.HTTP_BAD_REQUEST,
            const.INVALID_PARAMETER,
            "not a JSON body")
end

local action = args["BulkAction"]
if type(action) ~= "string" or #action == 0 then
    return utils.error_exit(
            ngx.HTTP_BAD_REQUEST,
            const.INVALID_PARAMETER,
            "illegal bulk action: " .. tostring(action))
end

action = string.lower(action)
local func = bulk_action_switch[action]
if func then
    local things = args["Things"]
    func(things)
else
    return utils.error_exit(
            ngx.HTTP_BAD_REQUEST,
            const.INVALID_PARAMETER,
            "unknown bulk action: " .. tostring(action))
end